import numpy
import six

from loopy.target import TargetBase
from loopy.target.c.codegen.expression import LoopyCCodeMapper


class AllToDouble(dict):
    """ This imitates a dict that maps everything to double and logs the requested keys """
    def __getitem__(self, key):
        self.__setitem__(key, numpy.float64)
        return numpy.float64


_registry = {'float32': 'float',
             'int32': 'int',
             'float64': 'double'}


class MyMapper(LoopyCCodeMapper):
    var_subst_map = {}


class DuneTarget(TargetBase):

    def get_or_register_dtype(self, names, dtype=None):
        return dtype

    def dtype_to_typename(self, dtype):
        # For now, we do this the simplest possible way
        assert(type(dtype) == numpy.dtype)
        return _registry[dtype.name]
#
#     def is_vector_dtype(self, dtype):
#         raise NotImplementedError()
#
#     def vector_dtype(self, base, count):
#         raise NotImplementedError()

    def get_expression_to_code_mapper(self, codegen_state):
        return MyMapper(codegen_state)

    def generate_code(self, kernel, codegen_state, impl_arg_info):
        from cgen import Block
        body = Block()

        from loopy.codegen.loop import set_up_hw_parallel_loops
        gen_code = set_up_hw_parallel_loops(kernel, 0, codegen_state)

        from cgen import Line
        body.append(Line())

        if isinstance(gen_code.ast, Block):
            body.extend(gen_code.ast.contents)
        else:
            body.append(gen_code.ast)

        return str(body), gen_code.implemented_domains

    def get_value_arg_decl(self, name, shape, dtype, is_written):
        assert shape == ()

        return "blubb"

    def get_global_arg_decl(self, name, shape, dtype, is_written):
        return "blubb"

    def emit_sequential_loop(self, codegen_state, iname, iname_dtype, static_lbound, static_ubound, inner):

        from loopy.codegen import wrap_in

        # Some of our loops are special as they use PDELab specific concepts.
        # Fortunately those loops are tied to specific inames.
        from dune.perftool.pdelab.quadrature import quadrature_iname
        if iname == quadrature_iname():
            from cgen import CustomLoop
            from dune.perftool.pdelab.quadrature import quadrature_loop_statement
            loop_stmt = quadrature_loop_statement()
            return wrap_in(CustomLoop,
                           "for({})".format(loop_stmt),
                           inner)

        # From here on it is the default implementation taken from loopys CTarget.
        ecm = codegen_state.expression_to_code_mapper

        from loopy.symbolic import aff_to_expr

        from pymbolic.mapper.stringifier import PREC_NONE
        from cgen import For

        return wrap_in(For,
                       "{} {} = {}".format(self.dtype_to_typename(iname_dtype),
                                           iname,
                                           ecm(aff_to_expr(static_lbound),
                                               PREC_NONE,
                                               "i",
                                               ),
                                           ),
                       "{} <= {}".format(iname,
                                         ecm(aff_to_expr(static_ubound),
                                             PREC_NONE,
                                             "i",
                                             ),
                                         ),
                       "++{}".format(iname),
                       inner,
                       )
